#define THIS_IS_MAIN
#include <iostream>
#include <iomanip>
#include "mps.h"
#include "model.h"
#include <boost/foreach.hpp>

#define foreach BOOST_FOREACH

using namespace std;

//For N = 6, isotropic case
//Heisenberg GS Energy/(N-1) = -0.498715

int main()
{
    //Settings
    Real tau = 0.05;
    int Nstep = 5000;
    Real cutoff = 1.0E-20;
    Real Jz = 1.0, Jxy = 1.0;

    //Construct lattice system
    const int N = 6; //Six sites
    const int d = 2; //Spin 1/2
    IndexSiteSet ISS(N,d);

    //Construct Hamiltonian
    vector<ITensor> H = makeHeis(ISS,Jz,Jxy);

    //Construct initial state psi
    ITensor psi;
    for(int i = 1; i <= N; ++i) { psi.add_index(ISS.si(i)); }
    psi.dat = 0.0;

    psi(1,2,1,2,1,2) = 1.0; //Neel state

    //Perform imaginary time evol.
    for(int n = 1; n <= Nstep; ++n)
    {
        ITensor psip = psi;

        //Multiply by H
        foreach(ITensor h, H)
        {
            psip = h*psip;
        }
        psip.noprime();

        Real En = RDot(psip,psi);
        cout << "Step " << n << ": Energy/(N-1) = " << En/(N-1) << endl; 

        //|psi> = |psi> - tau*H|psi>
        psip *= -tau;
        psi = psi + tau*psip;

        //Normalize
        Real Norm2 = RDot(psi,psi);
        psi *= 1.0/sqrt(Norm2);

    }

    //Measure energy
    ITensor psip = psi;

    foreach(ITensor h, H)
    {
        psip = h*psip;
    }
    psip.noprime();

    Real En = RDot(psip,psi);
    cout << endl << "Energy/(N-1) after " << Nstep << " steps = " << En/(N-1) << endl << endl;

    //Entropy calculations
    vector<Real> entropies; entropies.push_back(0.0);
    for(int cut_bond = 1; cut_bond < N; ++cut_bond)
    {
        ITensor phi = psi, phip;
        cout << "cut_bond = " << cut_bond << endl;

        //Set up combiners
        Index LH(Site,"LH"),RH(Site,"RH");
        Combiner cLH(LH),cRH(RH);
        for(int i = 1; i <= N; ++i)
        {
            if(i <= cut_bond) cLH.addleft(LH,ISS.si(i));
            else              cRH.addleft(RH,ISS.si(i));
        }

        //Apply combiners
        phi = cLH * phi * cRH;

        //Calc reduced density matrix
        phip = phi.primesite();
        phi.mapindex(RH,RH.primed());

        ITensor rho = phi * conj(phip);

        int m = LH.m();
        Matrix Rho(m,m);
        rho.getMatrix11(LH,LH.primed(),Rho);

        Matrix U; Vector D;
        Rho *= -1;
        EigenValues(Rho,D,U);
        Rho *= -1; D *= -1;

        //Truncate
        Real truncerr = 0.0;
        int mp = D.Length();
        while(truncerr + D(mp) < cutoff*D(1) && mp > 1) truncerr += D(mp--);
        if(D(1) == 0.0) truncerr = 0.0;
        else            truncerr = truncerr/D(1);
        D.ReduceDimension(mp);

        cout << "For cut on bond " << cut_bond << ", density matrix spectrum is:" << endl;
        for(int i = 1; i <= D.Length(); ++i)
            cout << i << " " << D(i) << endl;

        Real SvN = 0.0;
        for(int i = 1; i <= D.Length(); ++i) 
        {
            if(D(i) > 1E-20) SvN += -D(i)*log(D(i));
        }

        cout << endl << "SvN = " << SvN << endl << endl;

        entropies.push_back(SvN);
    }

    cout << "Calculated entropies: " << endl;
    for(int i = 1; i < N; ++i) cout << i SP entropies[i] << endl;


    return 0;
}
